<?php

include_once( "strings.php" );
include_once(    "json.php" );
include_once(     "log.php" );
include_once(     "ssl.php" );
include_once(   "mysql.php" );
include_once(   "input.php" );

$args = isset( $argv ) ? $argv : array();

main( $args );

function main( $argv )
{
    $start = microtime( TRUE );

    NeoLog( "Starting new request:" . $_SERVER["REDIRECT_URL"] . urldecode( file_get_contents( 'php://input') ) );
    NeoLogIn();

    $environment = DetermineEnvironment         ( $argv                                                           );
                   CheckIfTunnelRequest         ( $environment                                                    );
    $sp_name     = TranslateSPName              ( $environment["redirect_url"]                                    );
    $db_domain   = DetermineDatabaseLookupDomain( $environment["server_name"]                                     );
    $db_info     = LookupDatabaseInfo           ( $db_domain                                                      );
    $db_version  = LookupDatabaseVersion        ( $db_info                                                        );
    $sp_call     = MapRequestToSPCall           ( $db_info,     $db_version, $sp_name, $environment["request"]    );
    $ret         = CallStoredProcedure          ( $db_info,     $db_version, $sp_name, $sp_call, $environment     );
                   SetSessionIDCookie           ( $sp_name,     $ret                                              );
    $response    = CreateResponse               ( $environment, $ret, ceil( (microtime( TRUE ) - $start) * 1000 ) );
                   OutputResponse               ( $sp_name,     $response                                         );
}

function DetermineEnvironment( $argv )
{
    NeoLog( "Determining environment" );
    NeoLogIn();

    $environment = null;

    if ( "cli" != php_sapi_name() )
    {
        NeoLog( "Determining envionrment from Apache" );
        $environment = DetermineEnvironmentFromApache();
    }
    else
    {
        NeoLog( "Determining envionrment from command-line arguments" );
        $environment = DetermineEnvironmentFromArguments( $argv );
    }

    if ( ("" == $environment["server_name"]) || ("" == $environment["redirect_url"]) )
    {
        NeoLog( "Invalid environment, no SERVER_NAME or REDIRECT_URL." );
        NeoLog( "Commandline usage:" );
        NeoLog( "php index.php --server-name api.example.com --redirect-url /api/users/example/ --request '{\"param1\":\"value\",...}'" );
        exit;
    }

    if ( !defined( "DB_USERNAME" ) ) define( "DB_USERNAME", "public" );
    if ( !defined( "DB_PASSWORD" ) ) define( "DB_PASSWORD", "public" );
    if ( !defined( "SSL_BASE"    ) ) define( "SSL_BASE", "/etc/mysql/ssl/" );

    NeoLog( "Configuration: username: " . DB_USERNAME . "; password: " . DB_PASSWORD . "; ssl_base: " . SSL_BASE );

    NeoLogOut();

    return $environment;
}

function DetermineEnvironmentFromApache()
{
    $server_name  = "";
    $redirect_url = "";
    $requested    = "";

    if ( isset( $_SERVER ) && array_key_exists( "SERVER_NAME", $_SERVER ) )
    {
        $server_name = $_SERVER["SERVER_NAME"];
    }

    if ( isset( $_SERVER ) && array_key_exists( "REDIRECT_URL", $_SERVER ) )
    {
        $redirect_url = CanonicalisePath( $_SERVER["REDIRECT_URL"] );
    }

    if ( isset( $_REQUEST ) )
    {
        $request = Input::FilterInput( $_REQUEST, null );
    }

    return array
    (
        "server_name"  => $server_name,
        "redirect_url" => $redirect_url,
        "request"      => $request
    );
}

function CanonicalisePath( $constant )
{
    if ( '.php' == substr( $constant, -4 ) )
    {
        $constant = dirname( $constant );
    }

    if ( '/' != substr( $constant, -1 ) )
    {
        $constant .= "/";
    }

    return $constant;
}

function DetermineEnvironmentFromArguments( $argv )
{
    $server_name  = "";
    $redirect_url = "";
    $request      = "";

    $next_is_server_name  = 0;
    $next_is_redirect_url = 0;
    $next_is_request      = 0;
    foreach( $argv as $arg )
    {
        if ( $next_is_server_name )
        {
            $server_name         = $arg;
            $next_is_server_name = 0;
        }
        else
        if ( $next_is_redirect_url )
        {
            $redirect_url         = $arg;
            $next_is_redirect_url = 0;
        }
        else
        if ( $next_is_request )
        {
            $request         = json_decode( $arg, TRUE );
            $next_is_request = 0;
        }
        else
        {
            switch ( $arg )
            {
            case "--server-name":
                $next_is_server_name = 1;
                break;

            case "--redirect-url":
                $next_is_redirect_url = 1;
                break;

            case "--request":
                $next_is_request = 1;
                break;
            }
        }
    }

    if ( !$request ) $request = array();

    return array
    (
        "server_name"  => $server_name,
        "redirect_url" => $redirect_url,
        "request"      => $request
    );
}

function CheckIfTunnelRequest( $environment )
{
    NeoLog( "Checking if tunnel request" );

    $sp_name = TranslateSPName( $environment["redirect_url"] );

    if ( "tunnel" != $sp_name )
    {
        //
        //  This is mysteriously not being executed...
        //

        NeoLog( "Checking if tunnel request - NO" );
    }
    else
    {
        NeoLog( "Performing tunnel" );
        NeoLogIn();
        Tunnel( $environment );
        NeoLogOut();
    }
}

function Tunnel( $environment )
{
    $request      = $environment['request'];
    $method       = array_get( $request, "wab_method"       );
    $host         = array_get( $request, "wab_host"         );
    $endpoint     = array_get( $request, "wab_endpoint"     );
    $accept       = array_get( $request, "wab_accept"       );
    $content_type = array_get( $request, "wab_content_type" );
    $content64    = array_get( $request, "wab_content64"    );

    NeoLog( "WAB tunnel debug: $host$endpoint" );

    if ( ! ($method && $host && $endpoint && $accept && $content_type) )
    {
        header( "Content-Type: text/plain" );

        echo "Error: incomplete headers"         . "\n";
        echo "+ wab_method:       $method"       . "\n";
        echo "+ wab_host:         $host"         . "\n";
        echo "+ wab_endpoint:     $endpoint"     . "\n";
        echo "+ wab_accept:       $accept"       . "\n";
        echo "+ wab_content_type: $content_type" . "\n";
        echo "+ wab_content64:    $content64"    . "\n";
    }
    else
    {
        $curl         = SetupCURL( $method, $host, $endpoint, $content_type, $content64, $accept );
        $response     = curl_exec ( $curl );
        $err          = curl_error( $curl );

        if ( $err )
        {
            header( "Content-Type: text/plain" );
            echo $err;
        }
        else
        {
            header( "Content-Type: $accept" );
            echo $response;
        }
    }

    ob_flush();
    flush();
    exit;
}

function SetupCURL( $method, $host, $endpoint, $content_type, $content64, $accept )
{
    $curl = curl_init();

    $headers = array
    (
        "Accept: $accept",
        "Content-Type: $content_type",
        "Cache-Control: no-cache"
    );

    $url      = $host . $endpoint;
    $content  = base64_decode( $content64 );

    if ( "GET" == $method )
    {
        $url     = $url . "?" . $content;
        $content = "";
    }

    error_log( "WAB Tunnel: $url" );

    curl_setopt_array(
        $curl,
        array
        (
            CURLOPT_URL            => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING       => "",
            CURLOPT_MAXREDIRS      => 10,
            CURLOPT_TIMEOUT        => 30,
            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST  => $method,
            CURLOPT_POSTFIELDS     => $content,
            CURLOPT_HTTPHEADER     => $headers,
        )
    );

    return $curl;
}

function TranslateSPName( $redirect_url )
{
    NeoLog( "Generate Target Stored Procedure Name" );

    $bits = explode( '/', $redirect_url );

    if ( "" == $bits[0] ) $bits = array_slice( $bits, 1 );

    switch ( $bits[0] )
    {
    case 'api':
        $sp_name = trim( implode( '_', array_slice( $bits, 1 ) ), '_' );
        break;

    case 'auth':
        $sp_name = trim( implode( '_', array_slice( $bits, 0 ) ), '_' );
        break;

    case 'tunnel':
        $sp_name = 'tunnel';
        break;

    default:
        header( "Content-Type: text/plain" );
        echo "This is a Web API Bridge HTTP endpoint -- see https://www.webapibridge.org";
        exit;
    }

    return $sp_name;
}

function DetermineDatabaseLookupDomain( $server_name )
{
    if ( string_has_suffix( $server_name, ".local" ) )
    {
        $db_domain = "local-db." . str_replace( '.local', '', $server_name );
    }
    else
    {
        $db_domain = "db." . $server_name;
    }

    NeoLog( "Determining database lookup domain name: $db_domain" ); 

    return $db_domain;
}

function LookupDatabaseInfo( $db_domain )
{
    NeoLog( "Looking up database info from DNS" );
    NeoLogIn();

    $db_info = (object) array( "dbname" => null, "hosts" => null );

    if ( function_exists( "apcu_exists" ) && apcu_exists( $db_domain ) )
    {
        $db_info = apcu_fetch( $db_domain );
        $db_host = $db_info->hosts[0];

        NeoLog( "Caching: found cached: $db_domain --> $db_info->dbname@$db_host" );
    }
    else
    {
        $records = dns_get_record( $db_domain, DNS_TXT );
        $n       = count( $records );

        NeoLog( "$db_domain: returned $n records" );

        if ( 0 < $n )
        {
            $db_info = ExtractDBInfoFromDNSRecords( $records );
        }
        else
        {
            $db_info = ExtractDBInfoFromDBDomain( $db_domain );
        }

        if ( function_exists( "apcu_add" ) )
        {
            apcu_add( $db_domain, $db_info );
            NeoLog( "Caching: added to cache: $db_domain" );
        }
    }

    //  Following will abort if cannot retrieve database connection.
    CheckDatabaseConnection( $db_info );

    NeoLogOut();

    return $db_info;
}

function ExtractDBInfoFromDNSRecords( $records )
{
    $dbname = null;
    $hosts  = null;

    foreach ( $records as $record )
    {
        $host    = array_key_exists(    "host", $record ) ? $record[   "host"] : "";
        $class   = array_key_exists(   "class", $record ) ? $record[  "class"] : "";
        $ttl     = array_key_exists(     "ttl", $record ) ? $record[    "ttl"] : "";
        $type    = array_key_exists(    "type", $record ) ? $record[   "type"] : "";
        $txt     = array_key_exists(     "txt", $record ) ? $record[    "txt"] : "";

        $bits = explode( "@", $txt );
        
        switch ( count( $bits ) )
        {
        case 2:
            $dbname = $bits[0];
            $hosts  = explode( ",", $bits[1] );
            break;
        
        case 1:
            $hosts  = explode( ",", $bits[0] );
            break;
        }
    }

    return (object) array( "dbname" => $dbname, "hosts" => $hosts );
}

function ExtractDBInfoFromDBDomain( $db_domain )
{
    $truncated = str_replace( "www.", "", $db_domain );
    $truncated = str_replace( "api-", "", $truncated );
    $truncated = str_replace( "api.", "", $truncated );
    $bits      = explode( ".", $truncated );

    $n = count( $bits );

    for ( $i = 1; $i < $n; $i++ )
    {
        if ( ! IsTLD( $bits[$i], $i ) )
        {
            $dbname = $bits[$i];
            break;
        }
    }

    return (object) array( "dbname" => $dbname, "hosts" => array("localhost") );
}

function IsTLD( $domain_part, $level = 1 )
{
	switch ( $level )
	{
	case "1":
		switch ( $domain_part )
		{
		case "au":
		case "com":
		case "net";
		case "org";
		case "co":
		case "info":
		case "xyz":
		case "local":
			return true;
			
		default:
			return false;
		}
		break;

	case "2":
		switch ( $domain_part )
		{
		case "com":
		case "net";
		case "org";
		case "id":
		case "conf":
			return true;
			
		default:
			return false;
		}
		break;
	
	default:
		return false;
	}
}

function CheckDatabaseConnection( $db_info )
{
    if ( !defined( "DB_HOSTNAME" ) ) define( "DB_HOSTNAME", $db_info->hosts[0] );

    NeoLog( "Checking database connection to: " . $db_info->hosts[0] );
    NeoLogIn();

    $m = MySQLCreate();

    NeoLogOut();

    return TRUE;
}

function LookupDatabaseVersion( $db_info )
{
    NeoLog( "Determining target database version" );
    NeoLogIn();

	$db_version = "";
    $db_name    = "";
    $db_hosts   = implode( ', ', $db_info->hosts );
    $db_host    = $db_info->hosts[0];

    NeoLog( "Database hosts: " . $db_hosts ); 
    NeoLog( "Connecting to MySQL Host: " . $db_host ); 

    $mysql  = MySQLCreate();
    $tuples = MySQLInfo( $mysql, "SHOW DATABASES" );

    foreach( $tuples as $tuple )
    {
        $tmp = $tuple["Database"];

        if ( string_has_prefix( $tmp, $db_info->dbname ) )
        {
            $db_name = $tmp;
        }
    }
    $db_version = $db_name ? str_replace( $db_info->dbname, "", $db_name ) : $db_version;

    if ( $db_name )
    {
        NeoLog( "LookupDatabaseVersion: found: $db_name" );
    }
    else
    {
        NeoLog( "LookupDatabaseVersion: cound not find database for: " . $db_info->dbname );
        exit;
    }

    NeoLogOut();

    return $db_version;
}

function MapRequestToSPCall( $db_info, $db_version, $sp_name, $request )
{
    $sp_call = null;

    if ( "multiselect" != $sp_name )
    {
        NeoLog( "Determine Stored Procedure Call" );
        NeoLogIn();


        $db = $db_info->dbname . $db_version;

        //
        //  Create list of provisional sp names.
        //

        $proc_name = null;
        $provisional[] = $sp_name;
        if ( string_has_suffix( $sp_name, "_csv" ) )
        {
            $provisional[] = preg_replace( "/_csv\z/", "", $sp_name );
        }
        $provisional[] = $provisional[count($provisional)-1] . "_retrieve";

        //
        //  Use first procedure found that matches provisional name.
        //

        $result = null;
        foreach( $provisional as $prov )
        {
            $sql = "RetrieveParametersFor( '$db', '$prov' )";
            
            $result = MySQLFunction( $db, $sql );

            if ( null !== $result )
            {
                $proc_name = $prov;
                break;
            }
        }

        if ( !$proc_name )
        {
            NeoLog( "Could not find procedure for: " . $sp_name );
            exit;
        }
        else
        {
            $arguments = "( ";

            $parameters = ConvertParametersToDictionary( $result );

            foreach ( $parameters as $parameter )
            {
                // Strip of leading '$' and make lowercase.
                $parameter_name = str_replace( '$', '', $parameter );
                $value          = array_get( $request, $parameter_name );

                if ( ! $value ) $value = array_get( $request, strtolower( $parameter_name ) );

                $arguments     .= ("'" . $value . "', ");

                if ( defined( "VERBOSE" ) && VERBOSE )
                {
                    $len = strlen( $value );
                    error_log( "$parameter_name ($parameter) --> ($len) $value" );
                }
            }
            
            $arguments = substr( $arguments, 0, -2 );
            
            $arguments .= " )";

            $sp_call = $proc_name . $arguments;

            NeoLog( "SP Call: " . $sp_call );
        }

        NeoLogOut();
    }

    return $sp_call;
}

function ConvertParametersToDictionary( $string )
{
    $parameters = array();
    $bits       = explode( ",", trim( $string ) );
    
    foreach ( $bits as $parameter )
    {
        $parts = explode( " ", trim( $parameter ) );
        
        $parameters[] = $parts[0];

        if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $parts[0] . " " . $parts[1] );
    }
    
    return $parameters;
}

function CallStoredProcedure( $db_info, $db_version, $sp_name, $sp_call, $environment )
{
    $ret = null;

    NeoLog( "Call Stored Procedure" );
    NeoLogIn();

    $db = $db_info->dbname . $db_version;

    if ( "multiselect" == $sp_name )
    {
        $ret = Multiselect   ( $db, $environment );
    }
    else
    {
        $ret = MySQLProcedure( $db, $sp_call );
    }

    NeoLogOut();

    return $ret;
}

function Multiselect( $db, $environment )
{
    $set        = array();
    $kinds      = explode( ',', array_get( $environment['request'], "kinds" ) );
    $sid        = array_get( $environment['request'], "sid"    );
    $filter     = array_get( $environment['request'], "filter" );

    foreach ( $kinds as $kind )
    {
        $bits = explode( ":", $kind );
        switch ( count( $bits ) )
        {
        case 1:
            $id          = "";
            $select_name = $bits[0];
            break;

        case 2:
            $id          = explode( ':', $kind )[0];
            $select_name = explode( ':', $kind )[1];
            break;

        default:
            continue;
        }

        $select_call = "selects_$select_name( '$sid', '$id', '', '$filter' )";

        $tuples      = MySQLProcedure( $db, $select_call );

        if ( is_array( $tuples ) )
        {
            $set[] = (object) array( "id" => $id, "name" => $kind, "tuples" => $tuples, "call" => $select_call );
        }
        else
        {
            $set[] = (object) array( "id" => $id, "name" => $kind,  "error" => $tuples, "call" => $select_call );
        }
    }

    return $set;
}

function SetSessionIDCookie( $sp_name, $ret )
{
    NeoLog( "Existing sessionid: " . array_get( $_COOKIE, "sid" ) );

    if ( "auth_login" == $sp_name )
    {
        if ( 1 == count( $ret ) )
        {
            $tuple     = $ret[0];
            $sessionid = $tuple->sessionid;
            $cookie    = "Set-Cookie: sid=";
            $cookie    = $cookie . $sessionid;
            $cookie    = $cookie . "; path=/; HttpOnly";
        
            header( $cookie );

            NeoLog( "Setting Session ID Cookie: " . $sessionid );
        }
    }
}

function CreateResponse( $environment, $ret, $microseconds )
{
    $request = $environment["request"];

    $response              = array();
    $response['URL'      ] = $environment["redirect_url"];
    $response['error'    ] = !is_array( $ret ) ? $ret : "";
    $response['failover' ] = "FALSE";
    $response['hostname' ] = "";
    $response['message'  ] = "";
    $response['status'   ] = !is_array( $ret ) ? "ERROR" : "OK";
    $response['warning'  ] = "";
    $response['target_id'] = array_get( $request, "target_id" );
    $response['submit'   ] = array_get( $request, "submit"    );
    $response['limit'    ] = array_get( $request, "limit"     );
    $response['offset'   ] = array_get( $request, "offset"    );
    $response['microsecs'] = $microseconds;
    $response['results'  ] = is_array( $ret ) ? $ret : NULL;

    return (object) $response;
}

function OutputResponse( $sp_name, $response )
{
    NeoLog( "Output Response" );
    NeoLogIn();

    if ( string_has_suffix( $sp_name, "_csv" ) )
    {
        TranslateToCSV ( $response );
    }
    else
    {
        TranslateToJSON( $response );
    }

    NeoLogOut();
}

function TranslateToJSON( $results )
{
    NeoLog( "Translate to JSON" );

    ignore_user_abort( true );

    header( "Connection: close" );
    //header( "Content-Type: application/json" );
    header( "Content-Type: text/plain" );

    if ( ob_get_contents() ) ob_end_clean();

    ob_start();
    {
        echo \JSON4::Encode( 0, $results );
        echo "\n";
        $size = ob_get_length();
        header( "Content-Length: $size" );
        ob_end_flush();
        flush();
    }
}

function TranslateToCSV( $result )
{
    NeoLog( "Translate to CSV" );

    if ( "OK" == $result->status )
    {
        $content = "";

        $ds = ("true" == $this->getRequest( "double_space" ));

        if ( $map )
        {
            error_log( "Calling CSV::encode with mapped tuples." );

            $content = \CSV::encode( \CSV::MapValues( $map, $result->results ), true, $ds );
        }
        else
        {
            error_log( "Calling CSV::encode with unmapped tuples." );

            $content = \CSV::encode( $result->results, true, $ds );
        }

        $size = strlen( $content );

        if ( true )
        {
            error_log( "Content-Type: application/octet-stream" );
            error_log( "Content-Length: $size" );
            error_log( "Content-Disposition: attachment;filename=\"download\"" );
        }

        header( "Content-Type: application/octet-stream"                    );
        header( "Content-Length: $size"                                     );
        header( "Content-Disposition: attachment;filename=\"download.csv\"" );
        header( "Pragma: public"                                            );
        header( "Cache-Control: must-revalidate, post-check=0, pre-check=0" );
        echo $content;
        header( "Connection: close" );
    }
}

